#include "dynamic.h"

#include <gsl/gsl_blas.h>

void EF(fprint_dynamic)(FILE* stream, struct dynamic_type* descr, void* dynamic, unsigned int i) {
  gsl_vector* vec;
  gsl_matrix* mat;
  void* addr = (char*)(dynamic) + descr[i].offset;

  switch (descr[i].type) {
  case DOUBLE:
    fprintf(stream, "%s: %f", descr[i].name, *(double*)addr);

    break;
  case VECTOR:
    vec = *(gsl_vector**)addr;

    gsl_vector_fprintf(stream, vec, "%.3f");

    break;
  case MATRIX:
    mat = *(gsl_matrix**)addr;

    gsl_matrix_fprintf(stream, mat, "%.3f");

    break;
  case POSINT:
    fprintf(stream, "%u", *(unsigned int*)addr);

    break;
  default:
    assert(0);
  }
}

void EF(fprint_dynamics)(FILE* stream, struct dynamic_type* descr, unsigned int len, void* dynamic) {
  if (len > 0) {
    for(unsigned int i=0; i<len - 1; i++) {
      EF(fprint_dynamic)(stream, descr, dynamic, i);
      fprintf(stream, ", ");
    }
    EF(fprint_dynamic)(stream, descr, dynamic, len - 1);
  }
}

void EF(fprint)(FILE* stream, family* ef, param* param) {
  switch (param->type) {
  case SOURCE:
    fprintf(stream, "%s(lambda: ", ef->name);
    EF(fprint_dynamics)(stream, EF(source_descr), EF(source_len), param);
    if (EF(arguments_len) > 0)
      fprintf(stream, "; ");
    EF(fprint_dynamics)(stream, EF(arguments_descr), EF(arguments_len), ef);
    fprintf(stream, ")\n");
    break;

  case NATURAL:
    fprintf(stream, "%s(theta: ", ef->name);
    EF(fprint_dynamics)(stream, EF(natural_descr), EF(natural_len), param);
    if (EF(arguments_len) > 0)
      fprintf(stream, "; ");
    EF(fprint_dynamics)(stream, EF(arguments_descr), EF(arguments_len), ef);
    fprintf(stream, ")\n");
    break;

  case EXPECTATION:
    fprintf(stream, "%s(eta: ", ef->name);
    EF(fprint_dynamics)(stream, EF(expectation_descr), EF(expectation_len), param);
    if (EF(arguments_len) > 0)
      fprintf(stream, "; ");
    EF(fprint_dynamics)(stream, EF(arguments_descr), EF(arguments_len), ef);
    fprintf(stream, ")\n");
    break;

  default:
    assert(0);
  }
}

void EF(dynamic_alloc)(struct dynamic_type* descr, unsigned int len, unsigned int dim, param* p) {
  void* addr_p;

  for(unsigned int i=0; i<len; i++) {
    addr_p = (char*)(p) + descr[i].offset;

    switch (descr[i].type) {
    case DOUBLE:
      break;
    case VECTOR:
      *(gsl_vector**)addr_p = gsl_vector_alloc(dim);

      break;
    case MATRIX:
      *(gsl_matrix**)addr_p = gsl_matrix_alloc(dim, dim);

      break;
    default:
      assert(0);
    }
  }
}

void EF(dynamic_free)(struct dynamic_type* descr, unsigned int len, param* p) {
  void* addr_p;

  for(unsigned int i=0; i<len; i++) {
    addr_p = (char*)(p) + descr[i].offset;

    switch (descr[i].type) {
    case DOUBLE:
      break;
    case VECTOR:
      gsl_vector_free(*(gsl_vector**)addr_p);
      break;
    case MATRIX:
      gsl_matrix_free(*(gsl_matrix**)addr_p);
      break;
    default:
      assert(0);
    }
  }
}

param* EF(dynamic_copy)(struct dynamic_type* descr, unsigned int len, param* src, param* dest) {
  void* addr_src;
  void* addr_dest;

  gsl_vector* vec_src;
  gsl_vector* vec_dest;

  gsl_matrix* mat_src;
  gsl_matrix* mat_dest;

  for(unsigned int i=0; i<len; i++) {
    addr_src  = (char*)(src ) + descr[i].offset;
    addr_dest = (char*)(dest) + descr[i].offset;

    switch (descr[i].type) {
    case DOUBLE:
      *(double*)addr_dest = *(double*)addr_src;

      break;
    case VECTOR:
      vec_src  = *(gsl_vector**)addr_src;
      vec_dest = *(gsl_vector**)addr_dest;
      gsl_vector_memcpy(vec_dest, vec_src);

      break;
    case MATRIX:
      mat_src  = *(gsl_matrix**)addr_src;
      mat_dest = *(gsl_matrix**)addr_dest;

      gsl_matrix_memcpy(mat_dest, mat_src);

      break;
    default:
      assert(0);
    }
  }
  return dest;
}

param* EF(dynamic_add)(struct dynamic_type* descr, unsigned int len, param* p, param* q, param* dest) {
  void* addr_p;
  void* addr_q;
  void* addr_dest;

  gsl_vector* vec_p;
  gsl_vector* vec_q;
  gsl_vector* vec_dest;

  gsl_matrix* mat_p;
  gsl_matrix* mat_q;
  gsl_matrix* mat_dest;

  for(unsigned int i=0; i<len; i++) {
    addr_p    = (char*)(p   ) + descr[i].offset;
    addr_q    = (char*)(q   ) + descr[i].offset;
    addr_dest = (char*)(dest) + descr[i].offset;

    switch (descr[i].type) {
    case DOUBLE:
      *(double*)addr_dest = *(double*)addr_p + *(double*)addr_q;

      break;
    case VECTOR:
      vec_p    = *(gsl_vector**)addr_p;
      vec_q    = *(gsl_vector**)addr_q;
      vec_dest = *(gsl_vector**)addr_dest;

      gsl_vector_memcpy(vec_dest, vec_p);
      gsl_vector_add(vec_dest, vec_q);
      break;
    case MATRIX:
      mat_p    = *(gsl_matrix**)addr_p;
      mat_q    = *(gsl_matrix**)addr_q;
      mat_dest = *(gsl_matrix**)addr_dest;

      gsl_matrix_memcpy(mat_dest, mat_p);
      gsl_matrix_add(mat_dest, mat_q);

      break;
    default:
      assert(0);
    }
  }
  return dest;
}

param* EF(dynamic_minus)(struct dynamic_type* descr, unsigned int len, param* p, param* q, param* dest) {
  void* addr_p;
  void* addr_q;
  void* addr_dest;

  gsl_vector* vec_p;
  gsl_vector* vec_q;
  gsl_vector* vec_dest;

  gsl_matrix* mat_p;
  gsl_matrix* mat_q;
  gsl_matrix* mat_dest;

  for(unsigned int i=0; i<len; i++) {
    addr_p    = (char*)(p   ) + descr[i].offset;
    addr_q    = (char*)(q   ) + descr[i].offset;
    addr_dest = (char*)(dest) + descr[i].offset;

    switch (descr[i].type) {
    case DOUBLE:
      *(double*)addr_dest = *(double*)addr_p - *(double*)addr_q;

      break;
    case VECTOR:
      vec_p    = *(gsl_vector**)addr_p;
      vec_q    = *(gsl_vector**)addr_q;
      vec_dest = *(gsl_vector**)addr_dest;

      gsl_vector_memcpy(vec_dest, vec_p);
      gsl_vector_sub(vec_dest, vec_q);

      break;
    case MATRIX:
      mat_p    = *(gsl_matrix**)addr_p;
      mat_q    = *(gsl_matrix**)addr_q;
      mat_dest = *(gsl_matrix**)addr_dest;

      gsl_matrix_memcpy(mat_dest, mat_p);
      gsl_matrix_sub(mat_dest, mat_q);

      break;
    default:
      assert(0);
    }
  }
  return dest;
}

param* EF(dynamic_zero)(struct dynamic_type* descr, unsigned int len, param* p) {
  void* addr_p;
  gsl_vector* vec_p;
  gsl_matrix* mat_p;

  for(unsigned int i=0; i<len; i++) {
    addr_p    = (char*)(p   ) + descr[i].offset;

    switch (descr[i].type) {
    case DOUBLE:
      *(double*)addr_p = 0;

      break;
    case VECTOR:
      vec_p    = *(gsl_vector**)addr_p;

      gsl_vector_set_zero(vec_p);

      break;
    case MATRIX:
      mat_p    = *(gsl_matrix**)addr_p;

      gsl_matrix_set_zero(mat_p);

      break;
    default:
      assert(0);
    }
  }
  return p;
}

param* EF(dynamic_scale)(struct dynamic_type* descr, unsigned int len, param* p, double scale, param* dest) {
  void* addr_p;
  void* addr_dest;

  gsl_vector* vec_p;
  gsl_vector* vec_dest;

  gsl_matrix* mat_p;
  gsl_matrix* mat_dest;

  for(unsigned int i=0; i<len; i++) {
    addr_p    = (char*)(p   ) + descr[i].offset;
    addr_dest = (char*)(dest) + descr[i].offset;

    switch (descr[i].type) {
    case DOUBLE:
      *(double*)addr_dest = *(double*)addr_p * scale;

      break;
    case VECTOR:
      vec_p    = *(gsl_vector**)addr_p;
      vec_dest = *(gsl_vector**)addr_dest;

      gsl_vector_memcpy(vec_dest, vec_p);
      gsl_vector_scale(vec_dest, scale);

      break;
    case MATRIX:
      mat_p    = *(gsl_matrix**)addr_p;
      mat_dest = *(gsl_matrix**)addr_dest;

      gsl_matrix_memcpy(mat_dest, mat_p);
      gsl_matrix_scale(mat_dest, scale);

      break;
    default:
      assert(0);
    }
  }
  return dest;
}

double EF(dynamic_scalar)(struct dynamic_type* descr, unsigned int len, param* p, param* q) {
  gsl_vector* vec_p;
  gsl_vector* vec_q;

  gsl_matrix* mat_p;
  gsl_matrix* mat_q;

  double scalar = 0;

  for(unsigned int i=0; i<len; i++) {
    void* addr_p    = (char*)(p   ) + descr[i].offset;
    void* addr_q    = (char*)(q   ) + descr[i].offset;

    switch (descr[i].type) {
    case DOUBLE:
      scalar += *(double*)addr_p * *(double*)addr_q;
      break;

    case VECTOR:
      vec_p    = *(gsl_vector**)addr_p;
      vec_q    = *(gsl_vector**)addr_q;

      double ddot = 0;
      gsl_blas_ddot(vec_p, vec_q, &ddot);
      scalar += ddot;

      break;
    case MATRIX:
      mat_p    = *(gsl_matrix**)addr_p;
      mat_q    = *(gsl_matrix**)addr_q;

      size_t dim = mat_p->size1;
      gsl_matrix* tmp = gsl_matrix_alloc(dim, dim);
      /* tmp <- mat_p^T * mat_q */
      gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, mat_p, mat_q, 0.0, tmp);

      /* diag <- diag(tmp) */
      gsl_vector_view diag = gsl_matrix_diagonal(tmp);

      /* trace */
      for(size_t i=0; i<dim; i++) {
        scalar += gsl_vector_get(&diag.vector, i);
      }
      gsl_matrix_free(tmp);
      break;
    default:
      assert(0);
    }
  }
  return scalar;
}

